/*	Cory McWilliams
		mcwilc@rpi.edu
		10/2/2003 */

/*	Note:
		
		I created this vertex struct and all of the helpful functions I'm using with it
		a long time ago.  They're proving to be very useful operations, though it's
		bothersome to force float[3] and vertex3 to interact.
		
		Each function should be self-explanatory.
		
		I discovered Catmull-Rom splines by Googling, and while Bezier curves could do
		some things I've wanted to do in computer graphics, Catmull-Rom splines have the
		advantage of always going through all of my control points. */

#include "vertex.h"
#include <cmath>

float abs(const vertex &a)
{
	return static_cast<float>(sqrt(a.x * a.x + a.y * a.y + a.z * a.z));
}

vertex normalize(const vertex &a)
{
	float r = abs(a);
	vertex v ={a.x / r, a.y / r, a.z / r};
	return v;
}

vertex crossProduct(const vertex &a, const vertex &b)
{
	vertex v = {a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x};
	return v;
}

float dotProduct(const vertex &a, const vertex &b)
{
	return a.x * b.x + a.y * b.y + a.z * b.z;
}

vertex3 operator+(vertex3 const &a, vertex3 const &b)
{
	vertex3 v = {a.x + b.x, a.y + b.y, a.z + b.z};
	return v;
}

vertex3 operator-(vertex3 const &a, vertex3 const &b)
{
	vertex3 v = {a.x - b.x, a.y - b.y, a.z - b.z};
	return v;
}

vertex3 &operator+=(vertex3 &a, vertex3 const &b)
{
	return a = a + b;
}

vertex3 operator/(vertex3 const &a, float b)
{
	vertex3 v = {a.x / b, a.y / b, a.z / b};
	return v;
}

vertex3 &operator/=(vertex3 &a, float b)
{
	return a = a / b;
}

vertex3 Vertex3(float x, float y, float z)
{
	vertex3 v = {x, y, z};
	return v;
}

vertex3 operator*(float a, const vertex3 &r)
{
	vertex3 v = {a * r.x, a * r.y, a * r.z };
	return v;
}

vertex3 operator*(const vertex3 &r, float a)
{
	vertex3 v = {a * r.x, a * r.y, a * r.z };
	return v;
}

float relativeComponent(const vertex &r, const vertex &x, float absx)
{
	return dotProduct(r, x) / absx;
}

float relativeVelocity(const vertex &PA, const vertex &PB, const vertex &VA, const vertex &VB)
{
	float absx = abs(PB - PA);
	float va = relativeComponent(VA, PA - PB, absx);
	float vb = relativeComponent(VB, PB - PA, absx);
	return vb + va;
}

vertex3 cubicSpline(const vertex3 &v0, const vertex3 &v1, const vertex3 &v2, const vertex3 &v3, float t)
{
	return (v3 - 3 * v2 + 3 * v1 - v0) * t * t * t + (3 * v2 - 6 * v1 + 3 * v0) * t * t + (3 * v1 - 3 * v0) * t + (v0);
}

vertex3 catmullRomSpline(const vertex3 &v1, const vertex3 &v2, const vertex3 &v3, const vertex3 &v4, float t)
{
	// Equation taken from http://www.mvps.org/directx/articles/catmull/.

	return 0.5f * ((2 * v2) + (-1.0f * v1 + v3) * t + (2.0f * v1 - 5.0f * v2 + 4.0f * v3 - v4) * t * t + (-1.0f * v1 + 3.0f * v2 - 3.0f * v3 + v4) * t * t * t);
}

vertex3 catmullRomGradient(const vertex3 &v1, const vertex3 &v2, const vertex3 &v3, const vertex3 &v4, float t)
{
	return 0.5f * ((-1.0f * v1 + v3) + 2.0f * (2.0f * v1 - 5.0f * v2 + 4.0f * v3 - v4) * t + 3.0f * (-1.0f * v1 + 3.0f * v2 - 3.0f * v3 + v4) * t * t);
}

void catmullRomSpline(float *points, int count, float t, float v[3]) {
	int segment = int(t * count);
	t = (t * count - segment);
	vertex3 a = {points[(segment*3)%(count*3)],
		points[(segment*3+1)%(count*3)],
		points[(segment*3+2)%(count*3)]};
	vertex3 b = {points[(segment*3+3)%(count*3)],
		points[(segment*3+4)%(count*3)],
		points[(segment*3+5)%(count*3)]};
	vertex3 c = {points[(segment*3+6)%(count*3)],
		points[(segment*3+7)%(count*3)],
		points[(segment*3+8)%(count*3)]};
	vertex3 d = {points[(segment*3+9)%(count*3)],
		points[(segment*3+10)%(count*3)],
		points[(segment*3+11)%(count*3)]};
	vertex3 e = catmullRomSpline(a, b, c, d, t);
	v[0] = e.x;
	v[1] = e.y;
	v[2] = e.z;
}

void catmullRomGradient(float *points, int count, float t, float v[3]) {
	int segment = int(t * count);
	t = (t * count - segment);
	vertex3 a = {points[(segment*3)%(count*3)],
		points[(segment*3+1)%(count*3)],
		points[(segment*3+2)%(count*3)]};
	vertex3 b = {points[(segment*3+3)%(count*3)],
		points[(segment*3+4)%(count*3)],
		points[(segment*3+5)%(count*3)]};
	vertex3 c = {points[(segment*3+6)%(count*3)],
		points[(segment*3+7)%(count*3)],
		points[(segment*3+8)%(count*3)]};
	vertex3 d = {points[(segment*3+9)%(count*3)],
		points[(segment*3+10)%(count*3)],
		points[(segment*3+11)%(count*3)]};
	vertex3 e = catmullRomGradient(a, b, c, d, t);
	v[0] = e.x;
	v[1] = e.y;
	v[2] = e.z;
}
